// (c) COPYRIGHT MIT, INRIA and Keio, 2007 and onwards.
// Please first read the full copyright statement in file COPYRIGHT.html

package org.w3c.mwi.mobileok.basic;

import java.text.ParseException;
import java.util.Date;
import java.util.Map;

/**
 * Implements {@link TestType#CACHING} test.
 */
final class CachingTest extends AbstractTestImplementation {

	private static final String CACHE_CONTROL_HEADER = "Cache-Control";
	private static final String DATE_HEADER = "Date";
	private static final String ETAG_HEADER = "ETag";
	private static final String EXPIRES_HEADER = "Expires";
	private static final String IF_MODIFIED_SINCE_HEADER = "If-Modified-Since";
	private static final String LAST_MODIFIED_HEADER = "Last-Modified";
	private static final String PRAGMA_HEADER = "Pragma";

	private static final String NO_CACHE_DIRECTIVE = "no-cache";
	private static final String NO_MAX_AGE_DIRECTIVE = "max-age=0";

	public TestType getTestType() {
		return TestType.CACHING;
	}

	public TestResult runTest(final TestContext context) {

		final TestResult result = new TestResult(getTestType());

		final Map<String, String> headers = context.getResourceContext().getHeaders();

		final String expiresHeader = headers.get(EXPIRES_HEADER);
		final String cacheControlHeader = headers.get(CACHE_CONTROL_HEADER);
		addHeaderInfo(EXPIRES_HEADER, expiresHeader, result);
		addHeaderInfo(CACHE_CONTROL_HEADER, cacheControlHeader, result);

		if (expiresHeader == null && cacheControlHeader == null) {
			result.setOutcome(TestOutcome.FAIL);
			result.setOutcomeMessage(EXPIRES_HEADER + " and " + CACHE_CONTROL_HEADER + " headers not present");
		}

		if (cacheControlHeader != null) {
			if (cacheControlHeader.contains(NO_CACHE_DIRECTIVE)) {
				result.addWarning(CACHE_CONTROL_HEADER + " header contains '" + NO_CACHE_DIRECTIVE +
				                  "', rendering the response uncacheable. " +
				                  "It is desirable to make content cacheable when possible.");
			}
			if (cacheControlHeader.contains(NO_MAX_AGE_DIRECTIVE)) {
				result.addWarning(CACHE_CONTROL_HEADER + " header contains '" + NO_MAX_AGE_DIRECTIVE +
				                  "', rendering the response uncacheable. " +
				                  "It is desirable to make content cacheable when possible.");
			}
		}

		final String pragmaHeader = headers.get(PRAGMA_HEADER);
		addHeaderInfo(PRAGMA_HEADER, pragmaHeader, result);

		if (pragmaHeader != null && pragmaHeader.contains(NO_CACHE_DIRECTIVE)) {
			result.addWarning(PRAGMA_HEADER + " header contains '" + NO_CACHE_DIRECTIVE +
			                  "', rendering the response uncacheable. " +
			                  "It is desirable to make content cacheable when possible.");
		}

		final String dateHeader = headers.get(DATE_HEADER);
		addHeaderInfo(DATE_HEADER, dateHeader, result);

		if (expiresHeader != null && dateHeader != null) {
			try {
				final Date expiresDate = parseDateHeader(expiresHeader);
				try {
					final Date dateDate = parseDateHeader(dateHeader);
					if (!expiresDate.after(dateDate)) {
						result.addWarning(EXPIRES_HEADER + " header's value is not after " + DATE_HEADER +
						                  " header's value (" + expiresHeader + " vs. " +
						                  dateHeader + "). Response will not" +
						                  "be cacheable. It is desirable to make content cacheable when possible.");
					}
				} catch (ParseException pe2) {
					result.addWarning(DATE_HEADER + " header value is not in a valid format.");
				}
			} catch (ParseException pe) {
				result.addWarning(EXPIRES_HEADER +
				                  " header value is not in a valid format; response may not be cacheable." +
				                  "It is desirable to make content cacheable when possible.");
			}
		}

		final String lastModifiedHeader = headers.get(LAST_MODIFIED_HEADER);
		addHeaderInfo(LAST_MODIFIED_HEADER, lastModifiedHeader, result);
		if (lastModifiedHeader != null) {
			// TODO
		}

		final String eTagHeader = headers.get(ETAG_HEADER);
		addHeaderInfo(ETAG_HEADER, eTagHeader, result);
		if (eTagHeader != null) {
			// TODO
		}

		return result;
	}

}
