// (c) COPYRIGHT MIT, INRIA and Keio, 2007 and onwards.
// Please first read the full copyright statement in file COPYRIGHT.html

package org.w3c.mwi.mobileok.basic;

import org.apache.commons.httpclient.Header;
import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpMethod;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.codec.binary.Base64;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.io.IOException;
import java.net.URI;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.Charset;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * {@link HTTPResource} encapsulates information about the result of accessing a resource
 * via HTTP, including the headers returned and the body. It will also attempt to decode
 */
class HTTPResource {

	private static final String USER_AGENT_HEADER = "User-Agent";
	private static final String ACCEPT_HEADER = "Accept";
	private static final String ACCEPT_CHARSET_HEADER = "Accept-Charset";
	private static final String MOBILEOK_BASIC_UA = "W3C mobileOK DDC (http://www.w3.org/2006/07/mobileok-ddc)";
	private static final String MOBILEOK_BASIC_ACCEPT_HEADER =
		"application/xhtml+xml,text/html;q=0.1,application/vnd.wap.xhtml+xml;q=0.1,text/css,image/jpeg,image/gif";



	private static final Charset US_ASCII_CHARSET = Charset.forName("US-ASCII");

	private static final Set<String> KNOWN_REQUEST_HEADERS;
	private static final Set<String> KNOWN_RESPONSE_HEADERS;
	static {
		final Set<String> common = new HashSet<String>(20);
		common.add("Allow");
		common.add("Cache-Control");
		common.add("Connection");
		common.add("Content-Encoding");
		common.add("Content-Language");
		common.add("Content-Length");
		common.add("Content-Location");
		common.add("Content-MD5");
		common.add("Content-Range");
		common.add("Content-Type");
		common.add("Date");
		common.add("Expires");
		common.add("Last-Modified");
		common.add("MIME-Version");
		common.add("Pragma");
		common.add("Trailer");
		common.add("Transfer-Encoding");
		common.add("Upgrade");
		common.add("Via");
		common.add("Warning");
		final Set<String> request = new HashSet<String>(20 + common.size());
		request.add("Accept");
		request.add("Accept-Charset");
		request.add("Accept-Encoding");
		request.add("Accept-Language");
		request.add("Authorization");
		request.add("Expect");
		request.add("From");
		request.add("Host");
		request.add("If-Match");
		request.add("If-Modified-Since");
		request.add("If-None-Match");
		request.add("If-Range");
		request.add("If-Unmodified-Since");
		request.add("Max-Forwards");
		request.add("Proxy-Authorization");
		request.add("Range");
		request.add("Referer");
		request.add("TE");
		request.add("User-Agent");
		request.addAll(common);
		final Set<String> response = new HashSet<String>(10 + common.size());
		response.add("Accept-Ranges");
		response.add("Age");
		response.add("ETag");
		response.add("Location");
		response.add("Proxy-Authenticate");
		response.add("Retry-After");
		response.add("Server");
		response.add("Vary");
		response.add("WWW-Authenticate");
		response.addAll(common);

		KNOWN_REQUEST_HEADERS = Collections.unmodifiableSet(request);
		KNOWN_RESPONSE_HEADERS = Collections.unmodifiableSet(response);
	}

	private final URI uri;
	private final Map<String, String> responseHeaders;
	private final byte[] rawBody;
	private final Document requestResponseDOM;

	HTTPResource(final URI uri) throws IOException, TestException {
		this(uri, Collections.<String, String>emptyMap());
	}

	HTTPResource(final URI uri, final Map<String, String> additionalRequestHeaders) throws IOException, TestException {
		if (uri == null) {
			throw new IllegalArgumentException("uri is null");
		}
		if (additionalRequestHeaders.containsKey(ACCEPT_HEADER) ||
			additionalRequestHeaders.containsKey(ACCEPT_CHARSET_HEADER) ||
			additionalRequestHeaders.containsKey(USER_AGENT_HEADER)) {
			throw new IllegalArgumentException("additionalRequestHeaders contains a reserved header");
		}
		this.uri = uri;
		final HttpMethod method = new GetMethod(uri.toString());
		if (!additionalRequestHeaders.isEmpty()) {
			for (final Map.Entry<String, String> entry : additionalRequestHeaders.entrySet()) {
				method.setRequestHeader(entry.getKey(), entry.getValue());
			}
		}
		method.setRequestHeader(USER_AGENT_HEADER, MOBILEOK_BASIC_UA);
		method.setRequestHeader(ACCEPT_HEADER, MOBILEOK_BASIC_ACCEPT_HEADER);
		method.setRequestHeader(ACCEPT_CHARSET_HEADER, "UTF-8");
		final HttpClient client = new HttpClient();
		final int httpStatus = client.executeMethod(method);
		this.rawBody = method.getResponseBody();
		this.responseHeaders = parseHTTPHeaders(method.getResponseHeaders());
		method.releaseConnection();
		this.requestResponseDOM =
			buildDOMRepresentation(uri, additionalRequestHeaders, httpStatus, responseHeaders, rawBody);
	}

	HTTPResource(final Integer httpStatus,
	             final Map<String, String> responseHeaders,
	             final byte[] rawBody) throws TestException {
		if (rawBody == null) {
			throw new IllegalArgumentException("rawBody is null");
		}
		this.uri = null;
		this.rawBody = rawBody;
		if (responseHeaders == null) {
			this.responseHeaders = Collections.emptyMap();
		} else {
			this.responseHeaders = responseHeaders;
		}
		this.requestResponseDOM = buildDOMRepresentation(null,
		                                                 Collections.<String, String>emptyMap(),
		                                                 httpStatus,
		                                                 responseHeaders,
		                                                 rawBody);
	}

	URI getURI() {
		return uri;
	}

	Map<String, String> getHeaders() {
		return responseHeaders;
	}

	byte[] getRawBody() {
		return rawBody;
	}

	Document getRequestResponseDOM() {
		return requestResponseDOM;
	}

	private static Map<String, String> parseHTTPHeaders(final Header[] responseHeaders) {
		final Map<String, String> httpHeaders = new HashMap<String, String>(responseHeaders.length);
		for (final Header header : responseHeaders) {
			if (!header.isAutogenerated()) {
				httpHeaders.put(header.getName(), header.getValue());
			}
		}
		return httpHeaders;
	}

	private static Document buildDOMRepresentation(final URI uri,
	                                               final Map<String, String> additionalRequestHeaders,
	                                               final Integer httpStatus,
	                                               final Map<String, String> responseHeaders,
	                                               final byte[] rawBody) throws TestException {

		final Document document = DOMUtils.newDOM();		

		final Element root = document.createElementNS(DOMUtils.NAMESPACE_RDF, "rdf:RDF");
		root.setAttribute("xmlns:http", DOMUtils.NAMESPACE_HTTP);
		root.setAttribute("xmlns:uri", DOMUtils.NAMESPACE_URI);
		root.setAttribute("xmlns:rfc822", DOMUtils.NAMESPACE_RFC822);
		document.appendChild(root);
		
		final Element getRequest = document.createElementNS(DOMUtils.NAMESPACE_HTTP, "http:GetRequest");
		if (uri != null) {
			root.appendChild(getRequest);
			DOMUtils.addElementWithText(document, getRequest, DOMUtils.NAMESPACE_URI, "uri:uri", uri.toString());
			DOMUtils.addElementWithText(document, getRequest, DOMUtils.NAMESPACE_HTTP, "http:version", "1.1");
			for (final Map.Entry<String, String> entry : additionalRequestHeaders.entrySet()) {
				addHeaderToDOM(KNOWN_REQUEST_HEADERS, document, getRequest, entry.getKey(), entry.getValue());
			}
			addHeaderToDOM(KNOWN_REQUEST_HEADERS, document, getRequest, ACCEPT_HEADER, MOBILEOK_BASIC_ACCEPT_HEADER);
			addHeaderToDOM(KNOWN_REQUEST_HEADERS, document, getRequest, ACCEPT_CHARSET_HEADER, "UTF-8");
			addHeaderToDOM(KNOWN_REQUEST_HEADERS, document, getRequest, USER_AGENT_HEADER, MOBILEOK_BASIC_UA);
		}

		final Element response = document.createElementNS(DOMUtils.NAMESPACE_HTTP, "http:Response");
		root.appendChild(response);
		if (httpStatus != null) {
			DOMUtils.addElementWithText(document,
										response,
										DOMUtils.NAMESPACE_HTTP,
										"http:responseCode",
										httpStatus.toString());
		}
		for (final Map.Entry<String, String> entry : responseHeaders.entrySet()) {
			addHeaderToDOM(KNOWN_RESPONSE_HEADERS, document, response, entry.getKey(), entry.getValue());
		}

		if (rawBody != null) {
			final Element body = document.createElementNS(DOMUtils.NAMESPACE_HTTP, "http:body");
			response.appendChild(body);
			final byte[] base64RawBody = Base64.encodeBase64(rawBody);
			final CharBuffer buffer;
			try {
				buffer = US_ASCII_CHARSET.newDecoder().decode(ByteBuffer.wrap(base64RawBody));
			} catch (CharacterCodingException cce) {
				// can't happen?
				throw new TestException(cce);
			}
			body.appendChild(document.createCDATASection(buffer.toString()));
		}

		return document;
	}

	private static void addHeaderToDOM(final Set<String> knownHeaders,
	                                   final Document document,
	                                   final Element parent,
	                                   final String name,
	                                   final String value) {
		if (knownHeaders.contains(name)) {
			DOMUtils.addElementWithText(document,
			                            parent,
			                            DOMUtils.NAMESPACE_HTTP,
			                            "http:" + name.toLowerCase(),
			                            value);
		} else {
			final Element additionalHeader = document.createElementNS(DOMUtils.NAMESPACE_HTTP, "http:additionalHeader");
			parent.appendChild(additionalHeader);
			final Element rfc822Field = document.createElementNS(DOMUtils.NAMESPACE_RFC822, "rfc822:Field");
			additionalHeader.appendChild(rfc822Field);
			DOMUtils.addElementWithText(document, rfc822Field, DOMUtils.NAMESPACE_RFC822, "rfc822:fieldName", name);
			DOMUtils.addElementWithText(document, rfc822Field, DOMUtils.NAMESPACE_RFC822, "rfc822:fieldValue", value);
		}
	}

}
