// (c) COPYRIGHT MIT, INRIA and Keio, 2007 and onwards.
// Please first read the full copyright statement in file COPYRIGHT.html

package org.w3c.mwi.mobileok.basic;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.util.ArrayList;
import java.util.EnumMap;
import java.util.List;
import java.util.Map;
import java.util.Date;
import java.text.SimpleDateFormat;
import java.net.URI;

/**
 * Encapsulates the results of all tests in the mobileOK Basic suite.
 */
public final class TestResults {

	private URI uri;
	private TestOutcome overallOutcome;
	private final List<TestResult> resultsList;
	private final Map<TestType, TestResult> resultsMap;
	private final List<String> globalWarnings;
	private final List<String> globalInfos;
	private Document intermediateResultsDOM;

	TestResults(final URI uri, final Document intermediateResultsDOM) {
		if (uri == null || intermediateResultsDOM == null) {
			throw new IllegalArgumentException("uri or intermediateResultsDOM is null");
		}
		this.uri = uri;
		overallOutcome = TestOutcome.PASS;
		resultsList = new ArrayList<TestResult>();
		resultsMap = new EnumMap<TestType, TestResult>(TestType.class);
		this.globalWarnings = new ArrayList<String>();
		this.globalInfos = new ArrayList<String>();
		this.intermediateResultsDOM = intermediateResultsDOM;
	}

	void addTestResult(final TestResult result) {
		if (result == null) {
			throw new IllegalArgumentException("result is null");
		}
		resultsList.add(result);
		resultsMap.put(result.getType(), result);
		if (result.getOutcome() != TestOutcome.PASS) {
			overallOutcome = TestOutcome.FAIL;
		}
	}

	void setOverallOutcome(final TestOutcome overallOutcome) {
		if (overallOutcome == null) {
			throw new IllegalArgumentException("overallOutcome is null");
		}
		this.overallOutcome = overallOutcome;
	}

	public URI getURI() {
		return uri;
	}

	public TestOutcome getOverallOutcome() {
		return overallOutcome;
	}

	public List<TestResult> getAllResults() {
		return resultsList;
	}

	public TestResult getResultForTest(final TestType type) {
		return resultsMap.get(type);
	}

	void addGlobalWarning(final String warning) {
		if (warning == null) {
			throw new IllegalArgumentException("warning is null");
		}
		globalWarnings.add(warning);
	}

	public List<String> getGlobalWarnings() {
		return globalWarnings;
	}

	void addGlobalInfo(final String info) {
		if (info == null) {
			throw new IllegalArgumentException("info is null");
		}
		globalInfos.add(info);
	}

	public List<String> getGlobalInfos() {
		return globalInfos;
	}

	public Document getIntermediateResultsDOM() {
		return intermediateResultsDOM;
	}

	public Document getTestResultsDOM() throws TestException {

		final Document document = DOMUtils.newDOM();

		final Element root = document.createElementNS(DOMUtils.NAMESPACE_RDF, "rdf:RDF");
		root.setAttribute("xmlns:dc", DOMUtils.NAMESPACE_DC);
		root.setAttribute("xmlns:dct", DOMUtils.NAMESPACE_DCT);		
		root.setAttribute("xmlns:earl", DOMUtils.NAMESPACE_EARL);
		root.setAttribute("xmlns:foaf", DOMUtils.NAMESPACE_FOAF);
		document.appendChild(root);


		final Element software = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:Software");
		software.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:ID", "#mobileOKBasicRI");
		root.appendChild(software);
		final Element softwareTitle = DOMUtils.addElementWithText(document,
		                                                          software,
		                                                          DOMUtils.NAMESPACE_DC,
		                                                          "dc:title",
		                                                          "mobileOK Basic Tests Reference Implementation");
		softwareTitle.setAttribute("xml:lang", "en");
		DOMUtils.addElementWithText(document,
		                            software,
		                            DOMUtils.NAMESPACE_FOAF,
		                            "foaf:homepage",
		                            "http://validator.w3.org/mobile");
		DOMUtils.addElementWithText(document, software, DOMUtils.NAMESPACE_DCT, "dct:hasVersion", "1.0");


		final Element webContent = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:WebContent");
		webContent.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:ID", "#mobileResource");
		root.appendChild(webContent);
		// TODO: EARL hasn't finalized what the attribute name should be:
		DOMUtils.addElementWithText(document, webContent, DOMUtils.NAMESPACE_RDF, "rdf:resource", uri.toString());
		final Element webContentDate =
			DOMUtils.addElementWithText(document,
		                                webContent,
	                                    DOMUtils.NAMESPACE_DC,
	                                    "dc:date",
	                                    new SimpleDateFormat("yyyy-MM-dd").format(new Date()));
		webContentDate.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:datatype", "http://www.w3.org/2001/XMLSchema#gDate");

		for (final TestResult testResult : resultsList) {

			final String testID = "test_" + testResult.getType().toString();
			final Element testCase = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:TestCase");
			testCase.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:ID", testID);
			root.appendChild(testCase);
			DOMUtils.addElementWithText(document,
			                            testCase,
			                            DOMUtils.NAMESPACE_DC,
			                            "dc:title",
			                            testResult.getType().toString());

			final Element assertion = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:Assertion");
			assertion.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:ID", "#assertion");
			root.appendChild(assertion);

			final Element assertedBy = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:assertedBy");
			assertedBy.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:resource", "#mobileOKBasicRI");
			assertion.appendChild(assertedBy);

			final Element subject = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:subject");
			subject.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:resource", "#mobileResource");
			assertion.appendChild(subject);

			final Element mode = DOMUtils.addElementWithText(document,
															 assertion,
															 DOMUtils.NAMESPACE_EARL,
															 "earl:mode",
															 null);
			mode.setAttributeNS(DOMUtils.NAMESPACE_RDF,
								"rdf:resource",
								"http://www.w3.org/WAI/ER/EARL/nmg-strawman#automatic");

			final Element test = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:test");
			test.setAttributeNS(DOMUtils.NAMESPACE_RDF, "rdf:resource", '#' + testID);
			assertion.appendChild(test);

			final Element result = document.createElementNS(DOMUtils.NAMESPACE_EARL, "earl:result");
			assertion.appendChild(result);
			final Element validity =
				DOMUtils.addElementWithText(document, result, DOMUtils.NAMESPACE_EARL, "earl:validity", null);
			validity.setAttributeNS(
				DOMUtils.NAMESPACE_RDF, "rdf:resource", validityForOutcome(testResult.getOutcome()));
		}

		return document;
	}

	private static String validityForOutcome(final TestOutcome outcome) {
		switch (outcome) {
			case PASS:
				return "http://www.w3.org/WAI/ER/EARL/nmg-strawman#pass";
			case FAIL:
				return "http://www.w3.org/WAI/ER/EARL/nmg-strawman#fail";
		}
		return null; // can't get here?
	}

	void clear() {
		overallOutcome = TestOutcome.PASS;
		resultsList.clear();
		resultsMap.clear();
		globalWarnings.clear();
		globalInfos.clear();
		intermediateResultsDOM = null;
	}

	@Override
	public String toString() {
		final StringBuilder result = new StringBuilder(8192);
		result.append(overallOutcome);
		result.append(AbstractTestImplementation.LINE_SEPARATOR);
		if (!globalWarnings.isEmpty()) {
			result.append("GLOBAL WARNINGS:");
			result.append(AbstractTestImplementation.LINE_SEPARATOR);
			for (final String warning : globalWarnings) {
				result.append('\t');
				result.append(warning);
				result.append(AbstractTestImplementation.LINE_SEPARATOR);
			}
		}
		if (!globalInfos.isEmpty()) {
			result.append("GLOBAL INFO:");
			result.append(AbstractTestImplementation.LINE_SEPARATOR);
			for (final String warning : globalInfos) {
				result.append('\t');
				result.append(warning);
				result.append(AbstractTestImplementation.LINE_SEPARATOR);
			}
		}
		for (final TestResult testResult : resultsList) {
			result.append(testResult);
		}
		return result.toString();
	}

}
