// (c) COPYRIGHT MIT, INRIA and Keio, 2007 and onwards.
// Please first read the full copyright statement in file COPYRIGHT.html

package org.w3c.mwi.mobileok.basic;

import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ThreadFactory;

/**
 * An instance of {@link Tester} will conduct tests on a single resource.
 */
public final class Tester {

	private static final byte[] NO_BODY = new byte[0];
	private static final ExecutorService testExecutorService = Executors.newCachedThreadPool(new TestThreadFactory());

	private final URI uri;
	private final Map<String, String> httpHeaders;
	private final byte[] rawBody;
	private TestContext context;

	public Tester(final File file) throws TestException {
		this(fileToURI(file));
	}

	/**
	 * Helper method -- needed to get {@link #Tester(File)} to compile.
	 */
	private static URI fileToURI(final File file) throws TestException {
		if (file == null) {
			throw new IllegalArgumentException("file is null");
		}
		try {
			return new URI("file://" + file.getAbsolutePath());
		} catch (URISyntaxException use) {
			throw new TestException(use);
		}
	}

	public Tester(final URI uri) {
		this.uri = uri;
		this.httpHeaders = Collections.emptyMap();
		this.rawBody = NO_BODY;
	}

	public Tester(final Map<String, String> httpHeaders, final byte[] rawBody) {
		this.uri = null;
		if (httpHeaders == null) {
			this.httpHeaders = Collections.emptyMap();
		} else {
			this.httpHeaders = httpHeaders;
		}
		this.rawBody = rawBody == null ? NO_BODY : rawBody;
	}

	public TestResults runTests() throws TestException {
		try {
			createContext();
		} catch (IOException ioe) {
			return createFailureResults(ioe);
		}
		final TestResults results = context.getTestResults();
		final TestType[] testTypes = TestType.values();
		final List<Callable<Void>> testsToRun = new ArrayList<Callable<Void>>(testTypes.length);
		for (final TestType type : testTypes) {
			testsToRun.add(new TestCallable(type));
		}
		try {
			final List<Future<Void>> resultFutures = testExecutorService.invokeAll(testsToRun);
			for (final Future<Void> future : resultFutures) {
				future.get();
			}
		} catch (InterruptedException ie) {
			throw new TestException("Testing thread unexpectedly interrupted", ie);
		} catch (ExecutionException ee) {
			throw new TestException("Unexpected exception while running tests", ee);
		}
		return results;
	}

	public TestResults runTest(final TestType type) throws TestException {
		try {
			createContext();
		} catch (IOException ioe) {
			return createFailureResults(ioe);
		}
		final TestImplementation implementation = TestType.getImplementation(type);
		final TestResults results = context.getTestResults();
		results.addTestResult(implementation.runTest(context));
		return results;
	}

	private void createContext() throws IOException, TestException {
		if (context == null) {
			final HTTPTextResource textResource;
			if (uri == null) {
				textResource = new HTTPTextResource(null, httpHeaders, rawBody);
			} else {
				textResource = new HTTPTextResource(uri);
			}
			context = new TestContext(textResource);
		} else {
			context.resetResults();
		}
	}

	private TestResults createFailureResults(final Exception e) throws TestException {
		final TestResults testResults = new TestResults(uri, DOMUtils.newDOM());
		testResults.setOverallOutcome(TestOutcome.FAIL);
		testResults.addGlobalWarning(e.getMessage());
		return testResults;
	}

	public static void main(final String... args) throws Exception {
		if (args.length < 1) {
			throw new IllegalArgumentException("usage: Tester <URI>");
		}

		final Tester tester = new Tester(new URI(args[0]));
		final TestResults results = tester.runTests();
		System.out.println(results);

		final TransformerFactory tFactory = TransformerFactory.newInstance();
		final StreamResult result = new StreamResult(System.out);
		tFactory.newTransformer().transform(new DOMSource(results.getIntermediateResultsDOM()), result);
		tFactory.newTransformer().transform(new DOMSource(results.getTestResultsDOM()), result); 		
	}

	private static final class TestThreadFactory implements ThreadFactory {
		private final ThreadGroup threadGroup;
		private TestThreadFactory() {
			threadGroup = new ThreadGroup("mobileOK Basic Test ThreadGroup");
			threadGroup.setDaemon(true);
		}
		public Thread newThread(final Runnable runnable) {
			final Thread thread = new Thread(threadGroup, runnable);
			thread.setDaemon(true);
			return thread;
		}
	}

	private final class TestCallable implements Callable<Void> {
		private final TestType type;
		private TestCallable(final TestType type) {
			this.type = type;
		}
		public Void call() throws TestException {
			final TestImplementation implementation = TestType.getImplementation(type);
			final TestResults results = context.getTestResults();
			// A rare case in which this object is concurrently modified:
			final TestResult result = implementation.runTest(context);
			synchronized (results) {
				results.addTestResult(result);
			}
			return null;
		}
	}
}
